import asyncio
from decimal import Decimal
from typing import TYPE_CHECKING, Optional, Union

from ya_payment import RequestorApi, models

from golem.resources.allocation.allocation import Allocation
from golem.resources.base import _NULL, Resource, api_call_wrapper
from golem.resources.invoice.events import NewInvoice

if TYPE_CHECKING:
    from golem.node import GolemNode
    from golem.resources.agreement import Agreement  # noqa


class Invoice(Resource[RequestorApi, models.Invoice, "Agreement", _NULL, _NULL]):
    """A single invoice on the Golem Network.

    Usually created by a :any:`GolemNode` initialized with `collect_payment_events = True`.
    """

    def __init__(self, node: "GolemNode", id_: str, data: Optional[models.Invoice] = None):
        super().__init__(node, id_, data)
        asyncio.create_task(node.event_bus.emit(NewInvoice(self)))

    async def accept_full(self, allocation: Allocation) -> None:
        """Accept full invoice amount using a given :any:`Allocation`."""
        amount_str = (await self.get_data()).amount
        await self.accept(allocation, Decimal(amount_str))

    @api_call_wrapper()
    async def accept(self, allocation: Allocation, amount: Union[Decimal, float]) -> None:
        acceptance = models.Acceptance(
            total_amount_accepted=str(amount), allocation_id=allocation.id
        )
        await self.api.accept_invoice(self.id, acceptance)



import logging
from dataclasses import dataclass
from typing import Optional

from golem.resources.proposal.proposal import ProposalData

logger = logging.getLogger(__name__)


@dataclass
class LinearCoeffs:
    price_storage_gib: float = 0.0
    price_mem_gib: float = 0.0
    price_cpu_sec: float = 0.0
    price_duration_sec: float = 0.0
    price_initial: float = 0.0

    @classmethod
    def from_proposal_data(cls, proposal_data: ProposalData) -> Optional["LinearCoeffs"]:
        pricing_model = proposal_data.properties.get("golem.com.pricing.model")

        if pricing_model != "linear":
            logger.debug(
                "Proposal `%s` is not in `linear` pricing model, ignoring",
                proposal_data.proposal_id,
            )
            return None

        coeffs = proposal_data.properties.get("golem.com.pricing.model.linear.coeffs")
        usage_vector = proposal_data.properties.get("golem.com.usage.vector")

        if not (
            isinstance(coeffs, (list, tuple))
            and isinstance(usage_vector, (list, tuple))
            and len(coeffs) == len(usage_vector) + 1
        ):
            logger.debug(
                "Proposal `%s` linear pricing coeffs must be a sequence"
                "matching a length of usage vector + 1, ignoring",
                proposal_data.proposal_id,
            )

            return None

        usage_vector_mapping = {
            # `_` versions are deprecated but still used by providers
            "golem.usage.cpu_sec": "price_cpu_sec",
            "golem.usage.cpu-sec": "price_cpu_sec",
            "golem.usage.duration_sec": "price_duration_sec",
            "golem.usage.duration-sec": "price_duration_sec",
            "golem.usage.storage_gib": "price_storage_gib",
            "golem.usage.storage-gib": "price_storage_gib",
            "golem.usage.gib": "price_mem_gib",
        }
        build_dict = {"price_initial": float(coeffs[-1])}

        for usage, coeff in zip(usage_vector, coeffs):
            key = usage_vector_mapping.get(usage)
            if key is None:
                continue
            build_dict[key] = float(coeff)

        return cls(**build_dict)
